==> odxtools-2.5.1.tar.gz: odxtools-2.5.1/odxtools/diagdatadictionaryspec.py <==
# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass, field
from itertools import chain
from typing import TYPE_CHECKING, Optional, Any, Dict, List

from .utils import short_name_as_id
from .dataobjectproperty import (
    DataObjectProperty,
    DtcDop,
)
from .endofpdufield import EndOfPduField
from .envdata import EnvironmentData
from .envdatadesc import EnvironmentDataDescription
from .globals import logger
from .multiplexer import Multiplexer
from .nameditemlist import NamedItemList
from .structures import BasicStructure, create_any_structure_from_et
from .table import Table
from .units import UnitSpec
from .odxlink import OdxLinkId, OdxLinkDatabase, OdxDocFragment
from .specialdata import SpecialDataGroup, create_sdgs_from_et

if TYPE_CHECKING:
    from .diaglayer import DiagLayer

def _construct_named_item_list(iterable):
    return NamedItemList(short_name_as_id, iterable)


@dataclass()
class DiagDataDictionarySpec:
    dtc_dops: NamedItemList[DtcDop] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    data_object_props: NamedItemList[DataObjectProperty] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    structures: NamedItemList[BasicStructure] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    end_of_pdu_fields: NamedItemList[EndOfPduField] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    tables: NamedItemList[Table] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    env_data_descs: NamedItemList[EnvironmentDataDescription] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    env_datas: NamedItemList[EnvironmentData] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    muxs: NamedItemList[Multiplexer] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    unit_spec: Optional[UnitSpec] = None
    sdgs: List[SpecialDataGroup] = field(default_factory=list)

    def __post_init__(self):
        self._all_data_object_properties = _construct_named_item_list(
            chain(
                self.data_object_props,
                self.structures,
                self.end_of_pdu_fields,
                self.dtc_dops,
                self.tables,
            )
        )

        # The attributes are already initialized as (most likely normal) lists.
        # Convert them to NamedItemLists if necessary
        if not isinstance(self.dtc_dops, NamedItemList):
            self.dtc_dops = _construct_named_item_list(self.dtc_dops)

        if not isinstance(self.data_object_props, NamedItemList):
            self.data_object_props = _construct_named_item_list(self.data_object_props)

        if not isinstance(self.structures, NamedItemList):
            self.structures = _construct_named_item_list(self.structures)

        if not isinstance(self.end_of_pdu_fields, NamedItemList):
            self.end_of_pdu_fields = _construct_named_item_list(self.end_of_pdu_fields)

        if not isinstance(self.tables, NamedItemList):
            self.tables = _construct_named_item_list(self.tables)

        if not isinstance(self.env_data_descs, NamedItemList):
            self.env_data_descs = _construct_named_item_list(self.env_data_descs)

        if not isinstance(self.env_datas, NamedItemList):
            self.env_datas = _construct_named_item_list(self.env_datas)

        if not isinstance(self.muxs, NamedItemList):
            self.muxs = _construct_named_item_list(self.muxs)

    @staticmethod
    def from_et(et_element, doc_frags: List[OdxDocFragment]) \
            -> "DiagDataDictionarySpec":
        # Parse DOP-BASEs
        data_object_props = [
            DataObjectProperty.from_et(dop_element, doc_frags)
            for dop_element in et_element.iterfind("DATA-OBJECT-PROPS/DATA-OBJECT-PROP")
        ]

        structures = []
        for structure_element in et_element.iterfind("STRUCTURES/STRUCTURE"):
            structure = create_any_structure_from_et(structure_element, doc_frags)
            assert structure is not None
            structures.append(structure)

        end_of_pdu_fields = [
            EndOfPduField.from_et(eofp_element, doc_frags)
            for eofp_element in et_element.iterfind("END-OF-PDU-FIELDS/END-OF-PDU-FIELD")
        ]

        dtc_dops = []
        for dop_element in et_element.iterfind("DTC-DOPS/DTC-DOP"):
            dop = DataObjectProperty.from_et(dop_element, doc_frags)
            assert isinstance(dop, DtcDop)
            dtc_dops.append(dop)

        tables = [
            Table.from_et(table_element, doc_frags)
            for table_element in et_element.iterfind("TABLES/TABLE")
        ]

        env_data_descs = [
            EnvironmentDataDescription.from_et(env_data_desc_element, doc_frags)
            for env_data_desc_element in et_element.iterfind("ENV-DATA-DESCS/ENV-DATA-DESC")
        ]

        env_data_elements = chain(
            et_element.iterfind("ENV-DATAS/ENV-DATA"),
            # ODX 2.0.0 says ENV-DATA-DESC could contain a list of ENV-DATAS
            et_element.iterfind("ENV-DATA-DESCS/ENV-DATA-DESC/ENV-DATAS/ENV-DATA"),
        )
        env_datas = [
            EnvironmentData.from_et(env_data_element, doc_frags)
            for env_data_element in env_data_elements
        ]

        muxs = [
            Multiplexer.from_et(mux_element, doc_frags)
            for mux_element in et_element.iterfind("MUXS/MUX")
        ]

        if et_element.find("UNIT-SPEC") is not None:
            unit_spec = UnitSpec.from_et(et_element.find("UNIT-SPEC"), doc_frags)
        else:
            unit_spec = None

        # TODO: Parse different specs.. Which of them are needed?
        for (path, name) in [
            ("STATIC-FIELDS", "static fields"),
            ("DYNAMIC-LENGTH-FIELDS/DYNAMIC-LENGTH-FIELD", "dynamic length fields"),
            (
                "DYNAMIC-ENDMARKER-FIELDS/DYNAMIC-ENDMARKER-FIELD",
                "dynamic endmarker fields",
            ),
        ]:
            num = len(list(et_element.iterfind(path)))
            if num > 0:
                logger.info(f"Not implemented: Did not parse {num} {name}.")

        sdgs = create_sdgs_from_et(et_element.find("SDGS"), doc_frags)

        return DiagDataDictionarySpec(
            data_object_props=NamedItemList(short_name_as_id, data_object_props),
            structures=NamedItemList(short_name_as_id, structures),
            end_of_pdu_fields=NamedItemList(short_name_as_id, end_of_pdu_fields),
            dtc_dops=NamedItemList(short_name_as_id, dtc_dops),
            unit_spec=unit_spec,
            tables=NamedItemList(short_name_as_id, tables),
            env_data_descs=NamedItemList(short_name_as_id, env_data_descs),
            env_datas=NamedItemList(short_name_as_id, env_datas),
            muxs=NamedItemList(short_name_as_id, muxs),
            sdgs=sdgs,
        )

    def _build_odxlinks(self) -> Dict[OdxLinkId, Any]:
        odxlinks = {}
        for obj in chain(self.data_object_props,
                         self.structures,
                         self.end_of_pdu_fields,
                         self.tables,
                         ):
            odxlinks[obj.odx_id] = obj

        for obj in chain(self.data_object_props,
                         self.dtc_dops,
                         self.env_data_descs,
                         self.env_datas,
                         self.muxs,
                         self.sdgs,
                         self.structures,
                         self.end_of_pdu_fields,
                         self.tables,
                         ):
            odxlinks.update(obj._build_odxlinks())

        if self.unit_spec is not None:
            odxlinks.update(self.unit_spec._build_odxlinks())

        return odxlinks

    def _resolve_references(self,
                            parent_dl: "DiagLayer",
                            odxlinks: OdxLinkDatabase):

        for obj in self.data_object_props:
            obj._resolve_references(odxlinks)
        for obj in self.dtc_dops:
            obj._resolve_references(odxlinks)
        for obj in self.end_of_pdu_fields:
            obj._resolve_references(parent_dl, odxlinks)
        for obj in self.env_data_descs:
            obj._resolve_references(odxlinks)
        for obj in self.env_datas:
            obj._resolve_references(parent_dl, odxlinks)
        for obj in self.muxs:
            obj._resolve_references(odxlinks)
        for obj in self.sdgs:
            obj._resolve_references(odxlinks)
        for obj in self.structures:
            obj._resolve_references(parent_dl, odxlinks)
        for obj in self.tables:
            obj._resolve_references(odxlinks)

        if self.unit_spec:
            self.unit_spec._resolve_references(odxlinks)

    @property
    def all_data_object_properties(self):
        return self._all_data_object_properties

==> odxtools-3.0.0.tar.gz: odxtools-3.0.0/odxtools/diagdatadictionaryspec.py <==
# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH
from dataclasses import dataclass, field
from itertools import chain
from typing import TYPE_CHECKING, Optional, Any, Dict, List

from .utils import short_name_as_id
from .dataobjectproperty import (
    DataObjectProperty,
    DtcDop,
)
from .endofpdufield import EndOfPduField
from .envdata import EnvironmentData
from .envdatadesc import EnvironmentDataDescription
from .globals import logger
from .multiplexer import Multiplexer
from .nameditemlist import NamedItemList
from .structures import BasicStructure, create_any_structure_from_et
from .table import Table
from .units import UnitSpec
from .odxlink import OdxLinkId, OdxLinkDatabase, OdxDocFragment
from .specialdata import SpecialDataGroup, create_sdgs_from_et

if TYPE_CHECKING:
    from .diaglayer import DiagLayer

def _construct_named_item_list(iterable):
    return NamedItemList(short_name_as_id, iterable)


@dataclass
class DiagDataDictionarySpec:
    dtc_dops: NamedItemList[DtcDop]
    data_object_props: NamedItemList[DataObjectProperty]
    structures: NamedItemList[BasicStructure]
    end_of_pdu_fields: NamedItemList[EndOfPduField]
    tables: NamedItemList[Table]
    env_data_descs: NamedItemList[EnvironmentDataDescription]
    env_datas: NamedItemList[EnvironmentData]
    muxs: NamedItemList[Multiplexer]
    unit_spec: Optional[UnitSpec]
    sdgs: List[SpecialDataGroup]

    def __post_init__(self):
        self._all_data_object_properties = _construct_named_item_list(
            chain(
                self.data_object_props,
                self.structures,
                self.end_of_pdu_fields,
                self.dtc_dops,
                self.tables,
            )
        )

        # The attributes are already initialized as (most likely normal) lists.
        # Convert them to NamedItemLists if necessary
        if not isinstance(self.dtc_dops, NamedItemList):
            self.dtc_dops = _construct_named_item_list(self.dtc_dops)

        if not isinstance(self.data_object_props, NamedItemList):
            self.data_object_props = _construct_named_item_list(self.data_object_props)

        if not isinstance(self.structures, NamedItemList):
            self.structures = _construct_named_item_list(self.structures)

        if not isinstance(self.end_of_pdu_fields, NamedItemList):
            self.end_of_pdu_fields = _construct_named_item_list(self.end_of_pdu_fields)

        if not isinstance(self.tables, NamedItemList):
            self.tables = _construct_named_item_list(self.tables)

        if not isinstance(self.env_data_descs, NamedItemList):
            self.env_data_descs = _construct_named_item_list(self.env_data_descs)

        if not isinstance(self.env_datas, NamedItemList):
            self.env_datas = _construct_named_item_list(self.env_datas)

        if not isinstance(self.muxs, NamedItemList):
            self.muxs = _construct_named_item_list(self.muxs)

    @staticmethod
    def from_et(et_element, doc_frags: List[OdxDocFragment]) \
            -> "DiagDataDictionarySpec":
        # Parse DOP-BASEs
        data_object_props = [
            DataObjectProperty.from_et(dop_element, doc_frags)
            for dop_element in et_element.iterfind("DATA-OBJECT-PROPS/DATA-OBJECT-PROP")
        ]

        structures = []
        for structure_element in et_element.iterfind("STRUCTURES/STRUCTURE"):
            structure = create_any_structure_from_et(structure_element, doc_frags)
            assert structure is not None
            structures.append(structure)

        end_of_pdu_fields = [
            EndOfPduField.from_et(eofp_element, doc_frags)
            for eofp_element in et_element.iterfind("END-OF-PDU-FIELDS/END-OF-PDU-FIELD")
        ]

        dtc_dops = []
        for dop_element in et_element.iterfind("DTC-DOPS/DTC-DOP"):
            dop = DtcDop.from_et(dop_element, doc_frags)
            assert isinstance(dop, DtcDop)
            dtc_dops.append(dop)

        tables = [
            Table.from_et(table_element, doc_frags)
            for table_element in et_element.iterfind("TABLES/TABLE")
        ]

        env_data_descs = [
            EnvironmentDataDescription.from_et(env_data_desc_element, doc_frags)
            for env_data_desc_element in et_element.iterfind("ENV-DATA-DESCS/ENV-DATA-DESC")
        ]

        env_data_elements = chain(
            et_element.iterfind("ENV-DATAS/ENV-DATA"),
            # ODX 2.0.0 says ENV-DATA-DESC could contain a list of ENV-DATAS
            et_element.iterfind("ENV-DATA-DESCS/ENV-DATA-DESC/ENV-DATAS/ENV-DATA"),
        )
        env_datas = [
            EnvironmentData.from_et(env_data_element, doc_frags)
            for env_data_element in env_data_elements
        ]

        muxs = [
            Multiplexer.from_et(mux_element, doc_frags)
            for mux_element in et_element.iterfind("MUXS/MUX")
        ]

        if et_element.find("UNIT-SPEC") is not None:
            unit_spec = UnitSpec.from_et(et_element.find("UNIT-SPEC"), doc_frags)
        else:
            unit_spec = None

        # TODO: Parse different specs.. Which of them are needed?
        for (path, name) in [
            ("STATIC-FIELDS", "static fields"),
            ("DYNAMIC-LENGTH-FIELDS/DYNAMIC-LENGTH-FIELD", "dynamic length fields"),
            (
                "DYNAMIC-ENDMARKER-FIELDS/DYNAMIC-ENDMARKER-FIELD",
                "dynamic endmarker fields",
            ),
        ]:
            num = len(list(et_element.iterfind(path)))
            if num > 0:
                logger.info(f"Not implemented: Did not parse {num} {name}.")

        sdgs = create_sdgs_from_et(et_element.find("SDGS"), doc_frags)

        return DiagDataDictionarySpec(
            data_object_props=NamedItemList(short_name_as_id, data_object_props),
            structures=NamedItemList(short_name_as_id, structures),
            end_of_pdu_fields=NamedItemList(short_name_as_id, end_of_pdu_fields),
            dtc_dops=NamedItemList(short_name_as_id, dtc_dops),
            unit_spec=unit_spec,
            tables=NamedItemList(short_name_as_id, tables),
            env_data_descs=NamedItemList(short_name_as_id, env_data_descs),
            env_datas=NamedItemList(short_name_as_id, env_datas),
            muxs=NamedItemList(short_name_as_id, muxs),
            sdgs=sdgs,
        )

    def _build_odxlinks(self) -> Dict[OdxLinkId, Any]:
        odxlinks = {}
        for obj in chain(self.data_object_props,
                         self.structures,
                         self.end_of_pdu_fields,
                         self.tables,
                         ):
            odxlinks[obj.odx_id] = obj

        for obj in chain(self.data_object_props,
                         self.dtc_dops,
                         self.env_data_descs,
                         self.env_datas,
                         self.muxs,
                         self.sdgs,
                         self.structures,
                         self.end_of_pdu_fields,
                         self.tables,
                         ):
            odxlinks.update(obj._build_odxlinks())

        if self.unit_spec is not None:
            odxlinks.update(self.unit_spec._build_odxlinks())

        return odxlinks

    def _resolve_references(self,
                            parent_dl: "DiagLayer",
                            odxlinks: OdxLinkDatabase):

        for obj in self.data_object_props:
            obj._resolve_references(odxlinks)
        for obj in self.dtc_dops:
            obj._resolve_references(odxlinks)
        for obj in self.end_of_pdu_fields:
            obj._resolve_references(parent_dl, odxlinks)
        for obj in self.env_data_descs:
            obj._resolve_references(odxlinks)
        for obj in self.env_datas:
            obj._resolve_references(parent_dl, odxlinks)
        for obj in self.muxs:
            obj._resolve_references(odxlinks)
        for obj in self.sdgs:
            obj._resolve_references(odxlinks)
        for obj in self.structures:
            obj._resolve_references(parent_dl, odxlinks)
        for obj in self.tables:
            obj._resolve_references(odxlinks)

        if self.unit_spec:
            self.unit_spec._resolve_references(odxlinks)

    @property
    def all_data_object_properties(self):
        return self._all_data_object_properties
